/*
 * This file is part of Deliantra, the Roguelike Realtime MMORPG.
 * 
 * Copyright (©) 2005,2006,2007,2008 Marc Alexander Lehmann / Robin Redeker / the Deliantra team
 * Copyright (©) 2002,2007 Mark Wedel & Crossfire Development Team
 * Copyright (©) 1992,2007 Frank Tore Johansen
 * 
 * Deliantra is free software: you can redistribute it and/or modify it under
 * the terms of the Affero GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the Affero GNU General Public License
 * and the GNU General Public License along with this program. If not, see
 * <http://www.gnu.org/licenses/>.
 * 
 * The authors can be reached via e-mail to <support@deliantra.net>
 */

/* Started file Sept 1996 - initialization of gods in form of a
 * linked list -b.t.
 */

#include <global.h>
#include <living.h>
#include <spells.h>

/* add_god_to_list()- called only from init_gods */
static void
add_god_to_list (archetype *god_arch)
{
  if (!god_arch)
    {
      LOG (llevError, "ERROR: Tried to add null god to list!\n");
      return;
    }

  godlink *god = new godlink;

  // name=NULL;        /* how to describe the god to the player */
  // arch=NULL;        /* pointer to the archetype of this god */
  // id=0;             /* id of the god */
  // pantheon=NULL;    /* the group to which the god belongs (not implemented) */
  // next=NULL;        /* next god in this linked list */

  god->arch = god_arch;
  god->name = god_arch->object::name;

  // first_god->id MUST be the highest god number for other code to work currently
  if (!first_god)
    god->id = 1;
  else
    {
      god->id = first_god->id + 1;
      god->next = first_god;
    }

  first_god = god;

#ifdef DEBUG_GODS
  LOG (llevDebug, "Adding god %s (%d) to list\n", &god->name, god->id);
#endif
}

godlink *
get_rand_god ()
{
  godlink *god = first_god;
  int i;

  if (god)
    for (i = rndm (god->id) + 1; god; god = god->next)
      if (god->id == i)
        break;

  if (!god)
    LOG (llevError, "get_rand_god(): can't find a random god!\n");

  return god;
}

/* pntr_to_god_obj() - returns a pointer to the object
 * We need to be VERY carefull about using this, as we
 * are returning a pointer to the CLONE object. -b.t.
 */
object *
pntr_to_god_obj (godlink *godlnk)
{
  return godlnk && godlnk->arch
    ? godlnk->arch
    : 0;
}

/* init_gods() - this takes a look at all of the archetypes to find
 * the objects which correspond to the GODS (type GOD) */
void
init_gods ()
{
  LOG (llevDebug, "Initialising gods...\n");

  for_all_archetypes (at)
    if (at->type == GOD)
      add_god_to_list (at);

  LOG (llevDebug, "done.\n");
}

